/***************************************************************************\
    
    Copyright (c) 2001, 2002 David Schmenk
    
    All rights reserved.
    
    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, and/or sell copies of the Software, and to permit persons
    to whom the Software is furnished to do so, provided that the above
    copyright notice(s) and this permission notice appear in all copies of
    the Software and that both the above copyright notice(s) and this
    permission notice appear in supporting documentation.
    
    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
    OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
    HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
    INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
    FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
    NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
    WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
    
    Except as contained in this notice, the name of a copyright holder
    shall not be used in advertising or otherwise to promote the sale, use
    or other dealings in this Software without prior written authorization
    of the copyright holder.
    
\***************************************************************************/

/***************************************************************************\
*                                                                           *
*                 Starlight Xpress USB FX basic control macros              *
*                                                                           *
\***************************************************************************/
#ifndef _SX_USB_H_
#define _SX_USB_H_
/*
 * Vendor and product IDs.
 */
#define ECHO2_VENDOR_ID             0x4444
#define ECHO2_PRODUCT_ID            0x4220
/*
 * Vendor specific control commands.
 */
#define SX_USB_ECHO                 0
#define SX_USB_CLEAR_PIXELS         1
#define SX_USB_READ_PIXELS_DELAYED  2
#define SX_USB_READ_PIXELS          3
#define SX_USB_SET_TIMER            4
#define SX_USB_GET_TIMER            5
#define SX_USB_RESET                6
#define SX_USB_SET_CCD              7
#define SX_USB_GET_CCD              8
#define SX_USB_SET_GUIDE_PORT       9
#define SX_USB_WRITE_SERIAL_PORT    10
#define SX_USB_READ_SERIAL_PORT     11
#define SX_USB_SET_SERIAL           12
#define SX_USB_GET_SERIAL           13
#define SX_USB_CAMERA_MODEL         14
#define SX_USB_LOAD_EEPROM          15
/*
 * Caps bit definitions.
 */
#define SX_USB_CAPS_STAR2K          0x01
#define SX_USB_CAPS_EEPROM          0x04
#define SX_USB_CAPS_GUIDER          0x08
/*
 * Flags bit definitions.
 */
#define CCD_FLAGS_FIELD_ODD         0x01
#define CCD_FLAGS_FIELD_EVEN        0x02
#define CCD_FLAGS_FIELD_BOTH        (CCD_FLAGS_FIELD_ODD|CCD_FLAGS_FIELD_EVEN)
#define CCD_FLAGS_NOBIN_ACCUM       0x04
#define CCD_FLAGS_NOWIPE_FRAME      0x08
#define CCD_FLAGS_TDI               0x02
#define CCD_FLAGS_NOCLEAR_FRAME     0x04
/*
 * Function macros. Parameters marked as IN are passed as a value into
 * the macro.  A parameter marked as OUT is used to receive a value
 * in the macro.  A parameter marked IN_OUT is passed as a value and
 * receives a value.
 *
 * NOTE: The parameters marked OUT are NOT pointers -
 * these are macros and are expanded in-line.
 */
#define usbwrite write  // Change this to satisfy current environment
#define usbread  read   // Change this to satisfy current environment
/*
 * Open and close need to be implemented.
 */
#define sxfxOpen(/* OUT */ usbdev)          \
do {                                        \
    usbdev = 1;                             \
} while(0)
#define sxfxClose(/* IN_OUT */ usbdev)      \
do {                                        \
    usbdev = 0;                             \
} while(0)
/*
 * Clear pixels on given CCD.
 */
#define sxfxClearPixels(/* IN */ usbdev,    \
                        /* IN */ ccd,       \
                        /* IN */ flags)     \
do {                                        \
    unsigned char usbpkt[8];                \
    usbpkt[0]=USB_TYPE_VENDOR|USB_DIR_OUT;  \
    usbpkt[1]=SX_USB_CLEAR_PIXELS;          \
    usbpkt[2]=(flags);                      \
    usbpkt[3]=(flags)>>8;                   \
    usbpkt[4]=ccd;                          \
    usbpkt[5]=0;                            \
    usbpkt[6]=0;                            \
    usbpkt[7]=0;                            \
    usbwrite(usbdev,usbpkt,8);              \
} while(0)
/*
 * NOTE: Both ReadPixelsDelayed and ReadPixels must
 * be followed by a read of (width/xbin)*(height/ybin)
 * pixels (1 or 2 bytes).  You can read the pixel data
 * in blocks to provide progress feedback to the GUI.
 */
#define sxfxReadPixelsDelayed(/* IN */ usbdev, \
                              /* IN */ ccd,    \
                              /* IN */ flags,  \
                              /* IN */ xoffset,\
                              /* IN */ yoffset,\
                              /* IN */ width,  \
                              /* IN */ height, \
                              /* IN */ xbin,   \
                              /* IN */ ybin,   \
                              /* IN */ msec)   \
do {                                        \
    unsigned char usbpkt[22];               \
    usbpkt[0]=USB_TYPE_VENDOR|USB_DIR_OUT;  \
    usbpkt[1]=SX_USB_READ_PIXELS_DELAYED;   \
    usbpkt[2]=(flags);                      \
    usbpkt[3]=(flags)>>8;                   \
    usbpkt[4]=ccd;                          \
    usbpkt[5]=0;                            \
    usbpkt[6]=14;                           \
    usbpkt[7]=0;                            \
    usbpkt[8]=(xoffset);                    \
    usbpkt[9]=(xoffset)>>8;                 \
    usbpkt[10]=(yoffset);                   \
    usbpkt[11]=(yoffset)>>8;                \
    usbpkt[12]=(width);                     \
    usbpkt[13]=(width)>>8;                  \
    usbpkt[14]=(height);                    \
    usbpkt[15]=(height)>>8;                 \
    usbpkt[16]=xbin;                        \
    usbpkt[17]=ybin;                        \
    usbpkt[18]=(msec);                      \
    usbpkt[19]=(msec)>>8;                   \
    usbpkt[20]=(msec)>>16;                  \
    usbpkt[21]=(msec)>>24;                  \
    usbwrite(usbdev,usbpkt,22);             \
} while(0)
#define sxfxReadPixels(/* IN */ usbdev,     \
                       /* IN */ ccd,        \
                       /* IN */ flags,      \
                       /* IN */ xoffset,    \
                       /* IN */ yoffset,    \
                       /* IN */ width,      \
                       /* IN */ height,     \
                       /* IN */ xbin,       \
                       /* IN */ ybin)       \
do {                                        \
    unsigned char usbpkt[18];               \
    usbpkt[0]=USB_TYPE_VENDOR|USB_DIR_OUT;  \
    usbpkt[1]=SX_USB_READ_PIXELS_DELAYED;   \
    usbpkt[2]=(flags);                      \
    usbpkt[3]=(flags)>>8;                   \
    usbpkt[4]=ccd;                          \
    usbpkt[5]=0;                            \
    usbpkt[6]=14;                           \
    usbpkt[7]=0;                            \
    usbpkt[8]=(xoffset);                    \
    usbpkt[9]=(xoffset)>>8;                 \
    usbpkt[10]=(yoffset);                   \
    usbpkt[11]=(yoffset)>>8;                \
    usbpkt[12]=(width);                     \
    usbpkt[13]=(width)>>8;                  \
    usbpkt[14]=(height);                    \
    usbpkt[15]=(height)>>8;                 \
    usbpkt[16]=xbin;                        \
    usbpkt[17]=ybin;                        \
    usbwrite(usbdev,usbpkt,18);             \
} while(0)
/*
 * Reset camera.  Cancels all outstanding operations.
 */
#define sxfxReset(/* IN */ usbdev)          \
do {                                        \
    unsigned char usbpkt[8];                \
    usbpkt[0]=USB_TYPE_VENDOR|USB_DIR_OUT;  \
    usbpkt[1]=SX_USB_RESET;                 \
    usbpkt[2]=0;                            \
    usbpkt[3]=0;                            \
    usbpkt[4]=0;                            \
    usbpkt[5]=0;                            \
    usbpkt[6]=0;                            \
    usbpkt[7]=0;                            \
    usbwrite(usbdev,usbpkt,8);              \
} while(0)
/*
 * Note that the pixwidth and pixheight are floating
 * point values that represent the pixel size
 * in microns.
 */
#define sxfxGetCCD(/* IN  */ usbdev,        \
                   /* IN  */ ccd,           \
                   /* OUT */ width,         \
                   /* OUT */ height,        \
                   /* OUT */ pixwidth,      \
                   /* OUT */ pixheight,     \
                   /* OUT */ cmatrix,       \
                   /* OUT */ bpp,           \
                   /* OUT */ nserports,     \
                   /* OUT */ caps)          \
do {                                        \
    unsigned char usbpkt[17];               \
    usbpkt[0]=USB_TYPE_VENDOR|USB_DIR_IN;   \
    usbpkt[1]=SX_USB_GET_CCD;               \
    usbpkt[2]=0;                            \
    usbpkt[3]=0;                            \
    usbpkt[4]=ccd;                          \
    usbpkt[5]=0;                            \
    usbpkt[6]=17;                           \
    usbpkt[7]=0;                            \
    usbwrite(usbdev,usbpkt,8);              \
    usbread(usbdev,usbpkt,17);              \
    width=usbpkt[2]|(usbpkt[3]<<8);         \
    height=usbpkt[6]|(usbpkt[7]<<8);        \
    pixwidth=(usbpkt[8]|(usbpkt[9]<<8))/256.0;\
    pixheight=(usbpkt[10]|(usbpkt[11]<<8))/256.0);\
    cmatrix=usbpkt[12]|(usbpkt[13]<<8);     \
    bpp=usbpkt[14];                         \
    nserports=usbpkt[15];                   \
    caps=usbpkt[16];                        \
} while(0)
/*
 * Returns camera model.  GetCCD returns
 * more detailed information.
 */
#define sxfxGetModel(/* IN  */ usbdev,      \
                     /* OUT */ model)       \
do {                                        \
    unsigned char usbpkt[8];                \
    usbpkt[0]=USB_TYPE_VENDOR|USB_DIR_IN;   \
    usbpkt[1]=SX_USB_CAMERA_MODEL;          \
    usbpkt[2]=0;                            \
    usbpkt[3]=0;                            \
    usbpkt[4]=ccd;                          \
    usbpkt[5]=0;                            \
    usbpkt[6]=2;                            \
    usbpkt[7]=0;                            \
    usbwrite(usbdev,usbpkt,8);              \
    usbread(usbdev,usbpkt,2);               \
    model=usbpkt[0]|(usbpkt[1]<<8);         \
} while(0)
/*
 * Set the STAR2000 port.
 */
#define sxfxSetSTAR2K(/* IN */ usbdev,      \
                      /* IN */ star2k)      \
do {                                        \
    unsigned char usbpkt[8];                \
    usbpkt[0]=USB_TYPE_VENDOR|USB_DIR_OUT;  \
    usbpkt[1]=SX_USB_SET_STAR2K;            \
    usbpkt[2]=star2k;                       \
    usbpkt[3]=0;                            \
    usbpkt[4]=0;                            \
    usbpkt[5]=0;                            \
    usbpkt[6]=0;                            \
    usbpkt[7]=0;                            \
    usbwrite(usbdev,usbpkt,8);              \
} while(0)
/*
 * NOTE: string must be less than 64 characters long.
 */
#define sxfxWriteSerial(/* IN */ usbdev,    \
                        /* IN */ port,      \
                        /* IN */ string,    \
                        /* IN */ len)       \
do {                                        \
    unsigned char usbpkt[72],i;             \
    usbpkt[0]=USB_TYPE_VENDOR|USB_DIR_OUT;  \
    usbpkt[1]=SX_USB_SET_STAR2K;            \
    usbpkt[2]=1;                            \
    usbpkt[3]=0;                            \
    usbpkt[4]=port;                         \
    usbpkt[5]=0;                            \
    usbpkt[6]=len;                          \
    usbpkt[7]=0;                            \
    for(i=0;i<len;i++)                      \
        usbpkt[i+8]=string[i];              \
    usbwrite(usbdev,usbpkt,8+len);          \
} while(0)
/*
 * Returns available data and length.
 */
#define sxfxReadSerial(/* IN  */ usbdev,    \
                       /* IN  */ port,      \
                       /* OUT */ string,    \
                       /* OUT */ len)       \
do {                                        \
    unsigned char usbpkt[8];                \
    usbpkt[0]=USB_TYPE_VENDOR|USB_DIR_IN;   \
    usbpkt[1]=SX_USB_GET_SERIAL;            \
    usbpkt[2]=1;                            \
    usbpkt[3]=0;                            \
    usbpkt[4]=port;                         \
    usbpkt[5]=0;                            \
    usbpkt[6]=2;                            \
    usbpkt[7]=0;                            \
    usbwrite(usbdev,usbpkt,8);              \
    usbread(usbdev,usbpkt,2);               \
    if ((len=usbpkt[0])) {                  \
        usbpkt[0]=USB_TYPE_VENDOR|USB_DIR_IN;\
        usbpkt[1]=SX_USB_READ_SERIAL_PORT;  \
        usbpkt[2]=0;                        \
        usbpkt[3]=0;                        \
        usbpkt[4]=port;                     \
        usbpkt[5]=0;                        \
        usbpkt[6]=len;                      \
        usbpkt[7]=0;                        \
        usbwrite(usbdev,usbpkt,8);          \
        usbread(usbdev,string,len);         \
    }                                       \
} while(0)

#endif /* _SX_USB_H_ */


